"""Defines the functions that handle the community APIs for World of Warcraft

Functions:
    achievement(region_tag, achievement_id, release, locale)
    achievement_category(region_tag, category_id, release, locale)
    achievement_media(region_tag, achievement_id, release, locale)
    auction(region_tag, connected_realm_id, auction_house_id, release, locale)
    azerite_essence(region_tag, essence_id, release, locale)
    azerite_essence_media(region_tag, essence_id, release, locale)
    azerite_essence_search(region_tag, field_values, release, locale)
    conduit(region_tag, conduit_id, release, locale)
    connected_realm(region_tag, connected_realm_id, release, locale)
    connected_realm_search(region_tag, field_values, release, locale)
    covenant(region_tag, covenant_id, release, locale)
    covenant_media(region_tag, covenant_id, release, locale)
    creature(region_tag, creature_id, release, locale)
    creature_display_media(region_tag, display_id, release, locale)
    creature_family(region_tag, family_id, release, locale)
    creature_family_media(region_tag, family_id, release, locale)
    creature_search(region_tag, field_values, release, locale)
    creature_type(region_tag, type_id, release, locale)
    guild_crest_components_index(region_tag, release, locale)
    guild_crest_media(region_tag, emblem_id, release, locale)
    heirloom(region_tag, heirloom_id, release, locale)
    item(region_tag, item_id, release, locale)
    item_class(region_tag, class_id, release, locale)
    item_media(region_tag, item_id, release, locale)
    item_search(region_tag, field_values, release, locale)
    item_set(region_tag, set_id, release, locale)
    item_subclass(region_tag, class_id, subclass_id, release, locale)
    journal_encounter(region_tag, encounter_id, release, locale)
    journal_encounter_search(region_tag, field_values, release, locale)
    journal_expansion(region_tag, expansion_id, release, locale)
    journal_instance(region_tag, instance_id, release, locale)
    journal_instance_media(region_tag, instance_id, release, locale)
    media_search(region_tag, field_values, release, locale)
    modified_crafting(region_tag, release, locale)
    modified_crafting_category(region_tag, category_id, release, locale)
    modified_crafting_reagent_slot_type(region_tag, slot_type_id, release, locale)
    mount(region_tag, mount_id, release, locale)
    mount_search(region_tag, field_values, release, locale)
    mythic_keystone_affix(region_tag, affix_id, release, locale)
    mythic_keystone_affix_media(region_tag, affix_id, release, locale)
    mythic_keystone_dungeon(region_tag, dungeon_id, release, locale)
    mythic_keystone_index(region_tag, release, locale)
    mythic_keystone_leaderboard(region_tag, connected_realm_id, dungeon_id, period_id, release, locale)
    mythic_keystone_period(region_tag, period_id, release, locale)
    mythic_keystone_season(region_tag, season_id, release, locale)
    mythic_raid_leaderboard(region_tag, raid_name, faction, release, locale)
    pet(region_tag, pet_id, release, locale)
    pet_ability(region_tag, pet_ability_id, release, locale)
    pet_ability_media(region_tag, ability_id, release, locale)
    pet_media(region_tag, pet_id, release, locale)
    playable_class(region_tag, class_id, release, locale)
    playable_class_media(region_tag, class_id, release, locale)
    playable_race(region_tag, race_id, release, locale)
    playable_spec(region_tag, spec_id, release, locale)
    playable_spec_media(region_tag, spec_id, release, locale)
    power_type(region_tag, power_id, release, locale)
    profession(region_tag, profession_id, release, locale)
    profession_media(region_tag, profession_id, release, locale)
    profession_skill_tier(region_tag, profession_id, skill_tier_id, release, locale)
    pvp_leader_board(region_tag, season_id, pvp_bracket, release, locale)
    pvp_rewards_index(region_tag, season_id, release, locale)
    pvp_season(region_tag, season_id, release, locale)
    pvp_talent(region_tag, pvp_talent_id, release, locale)
    pvp_talent_slots(region_tag, class_id, release, locale)
    pvp_tier(region_tag, tier_id, release, locale)
    pvp_tier_media(region_tag, tier_id, release, locale)
    quest(region_tag, quest_id, release, locale)
    quest_area(region_tag, quest_area_id, release, locale)
    quest_category(region_tag, quest_category_id, release, locale)
    quest_type(region_tag, quest_type_id, release, locale)
    realm(region_tag, realm_slug, release, locale)
    realm_search(region_tag, field_values, release, locale)
    recipe(region_tag, recipe_id, release, locale)
    recipe_media(region_tag, recipe_id, release, locale)
    region(region_tag, region_req, release, locale)
    reputation_faction(region_tag, faction_id, release, locale)
    reputation_tier(region_tag, tier_id, release, locale)
    soulbind(region_tag, soulbind_id, release, locale)
    spell(region_tag, spell_id, release, locale)
    spell_media(region_tag, spell_id, release, locale)
    spell_search(region_tag, field_values, release, locale, )
    talent(region_tag, talent_id, release, locale)
    tech_talent(region_tag, talent_id, release, locale)
    tech_talent_media(region_tag, talent_id, release, locale)
    tech_talent_tree(region_tag, tree_id, release, locale)
    toy(region_tag, toy_id, release, locale)
    title(region_tag, title_id, release, locale)
    wow_token_index(region_tag, release, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from typing import Optional, Any, Dict, Union, Tuple

from battlenet_client.wow import utils as wow_utils
from battlenet_client import utils
from battlenet_client import exceptions
from battlenet_client.decorators import verify_region

__version__ = '3.0.0'
__author__ = 'David \'Gahd\' Couples'


@verify_region
def achievement_category(region_tag: str, category_id: Optional[Union[int, str]] = "index", *,
                         release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of achievement categories, or an achievement category by ID

    Args:
        region_tag (str): region_tag abbreviation
        category_id (int, optional): the achievement's category ID or None (default).
            None will retrieve the entire list of achievement categories
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/achievement-category/{category_id}"

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def achievement(region_tag: str, achievement_id: Optional[Union[int, str]] = "index", *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of achievements, or an achievement by ID

    Args:
        region_tag (str): region_tag abbreviation
        achievement_id (int, optional): the achievement ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/achievement/{achievement_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    return uri, params


@verify_region
def achievement_media(region_tag: str, achievement_id: int, *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for an achievement by ID.

    Args:
        region_tag (str): region_tag abbreviation
        achievement_id (int): the achievement ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/achievement/{achievement_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def auction(region_tag: str, connected_realm_id: int, *, auction_house_id: Optional[int] = None,
            release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns all active auctions for a connected realm.

    See the Connected Realm API for information about retrieving a list of connected realm IDs.
    Auction house data updates at a set interval. The value was initially set at 1 hour;
    however, it might change over time without notice.

    Depending on the number of active auctions on the specified connected realm, the response
    from this endpoint may be rather large, sometimes exceeding 10 MB.

    See the Connected Realm WoWUtils.api for information about retrieving a list of
    connected realm IDs.

    Auction house data updates at a set interval. The value was initially set
    at 1 hour; however, it might change over time without notice.

    Depending on the number of active auctions on the specified connected realm,
    the response from this game_data may be rather large, sometimes exceeding
    10 MB.

    Args:
        region_tag (str): region_tag abbreviation
        connected_realm_id (int): the id of the connected realm
        auction_house_id (int, optional): the ID of the auction house
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        exceptions.BNetReleaseError: when an AH ID is used for the retail

    Notes:
        Auction house functionality is not available for WoW 1.x (Vanilla Classic)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/connected-realm/{connected_realm_id}/"

    if release == "retail" and auction_house_id is None:
        uri += "auctions"

    if release != "retail" and not auction_house_id:
        uri += "auctions/index"

    if release != "retail" and auction_house_id:
        uri += f"auctions/{auction_house_id}"

    if release == "retail" and auction_house_id:
        raise exceptions.BNetReleaseError("Auction House ID not available for retail")

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def azerite_essence(region_tag: str, essence_id: Optional[Union[int, str]] = "index", *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of azerite essences, or an azerite essence by ID.

    Args:
        region_tag (str): region_tag abbreviation
        essence_id (int, optional): the Azerite essence ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/azerite-essence/{essence_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def azerite_essence_search(region_tag: str, field_values: Dict[str, Any], *,
                           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of azerite essences. For more detail see the search guide at:
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): search criteria, as key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{utils.api_host(region_tag)}/data/wow/search/azerite-essence"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    params.update(field_values)

    return uri, params


@verify_region
def azerite_essence_media(region_tag: str, essence_id: int, *,
                          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for an azerite essence by ID.

    Args:
        region_tag (str): region_tag abbreviation
        essence_id (int): the azerite essence ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = (
        f"{utils.api_host(region_tag)}/data/wow/media/azerite-essence/{essence_id}"
    )
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def connected_realm(region_tag: str, connected_realm_id: Optional[Union[int, str]] = "index", *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of connected realms, or a connected realm by ID.

    A connected realm is a collection of realms operating as one larger realm

    Args:
        region_tag (str): region_tag abbreviation
        connected_realm_id (int, optional): the ID of the connected realm
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/connected-realm/{connected_realm_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def connected_realm_search(region_tag: str, field_values: Dict[str, Any], *,
                           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of connected realms. For more detail see the search guide:
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): field/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/connected-realm"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }
    params.update(field_values)
    return uri, params


# noinspection DuplicatedCode
@verify_region
def covenant(region_tag: str, covenant_id: Optional[Union[int, str]] = "index", *,
             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of covenants, or a covenant by ID.

    Args:
        region_tag (str): region_tag abbreviation
        covenant_id (int, optional): the ID of the covenant or the default 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/covenant/{covenant_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def covenant_media(region_tag: str, covenant_id: int, *,
                   release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a covenant by ID.

    Args:
        region_tag (str): region_tag abbreviation
        covenant_id (int, optional): the covenant ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/covenant/{covenant_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def soulbind(region_tag: str, soulbind_id: Optional[Union[int, str]] = "index", *,
             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of soulbinds, or a soulbind by ID.

    Args:
        region_tag (str): region_tag abbreviation
        soulbind_id (int, optional): the ID of the soulbind or the word of 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/covenant/soulbind/{soulbind_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def conduit(region_tag: str, conduit_id: Optional[Union[int, str]] = "index", *,
            release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of conduits, or a conduit by ID.

    Args:
        region_tag (str): region_tag abbreviation
        conduit_id (int, optional): the ID of the conduit or the word of 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/covenant/conduit/{conduit_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def creature_family(region_tag: str, family_id: Optional[Union[int, str]] = "index", *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of creature families, or a creature family by ID.

    Args:
        region_tag (str): region_tag abbreviation
        family_id (int, optional): the creature family ID or the default 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/creature-family/{family_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def creature_type(region_tag: str, type_id: Optional[Union[int, str]] = "index", *,
                  release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of creature types, or a creature type by ID.

    Args:
        region_tag (str): region_tag abbreviation
        type_id (int, optional): the creature type ID or the default 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/creature-type/{type_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def creature(region_tag: str, creature_id: int, *,
             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns a creature by ID.

    Args:
        region_tag (str): region_tag abbreviation
        creature_id (int, optional): the creature ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/creature/{creature_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def creature_search(region_tag: str, field_values: Dict[str, Any], *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of creatures. For more detail see the search guide
     https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): matching criteria in key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/creature"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    params.update(field_values)
    return uri, params


@verify_region
def creature_display_media(region_tag: str, display_id: int, *,
                           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a creature display by ID.

    Args:
        region_tag (str): region_tag abbreviation
        display_id (int, optional): the creature display ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = (
        f"{utils.api_host(region_tag)}/data/wow/media/creature-display/{display_id}"
    )
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def creature_family_media(region_tag: str, family_id: int, *,
                          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a creature family by ID.

    Args:
        region_tag (str): region_tag abbreviation
        family_id (int, optional): the creature family ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = (
        f"{utils.api_host(region_tag)}/data/wow/media/creature-family/{family_id}"
    )
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def guild_crest_components_index(region_tag: str, release: Optional[str] = "retail", *,
                                 locale: Optional[str] = 'enus') -> Tuple:
    """Returns an index of guild crest media.

    Args:
        region_tag (str): region_tag abbreviation
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/guild-crest/index"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def guild_crest_media(region_tag: str, category: str, icon_id: int, *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a guild crest border by ID.

    Args:
        region_tag (str): region_tag abbreviation
        category (str): either 'border' or 'emblem'
        icon_id (int): the border ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    if category and category not in ('border', 'emblem'):
        raise exceptions.BNetValueError("Improper category")

    uri = f"{utils.api_host(region_tag)}/data/wow/media/guild-crest/{category}/{icon_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def heirloom(region_tag: str, heirloom_id: Optional[Union[str, int]] = "index", *,
             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    uri = f"{utils.api_host(region_tag)}/data/wow/heirloom/{heirloom_id}"

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    return uri, params


@verify_region
def item_class(region_tag: str, class_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of item classes, or an item class by ID.

    Args:
        region_tag (str): region_tag abbreviation
        class_id (int, optional): item class ID or the default 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/item-class/{class_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def item_set(region_tag: str, set_id: Optional[Union[int, str]] = "index", *,
             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    """Returns an index of item sets, or an item set by ID.

    Args:
        region_tag (str): region_tag abbreviation
        set_id (int, optional): the item class ID or the default 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/item-set/{set_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def item_subclass(region_tag: str, class_id: int, subclass_id: int, *,
                  release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an item subclass by ID.

    Args:
        region_tag (str): region_tag abbreviation
        class_id (int): the item class ID
        subclass_id (int, optional): the item's subclass ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/item-class/{class_id}/item-subclass/{subclass_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def item(region_tag: str, item_id: int, *,
         release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an item by ID.

    Args:
        region_tag (str): region_tag abbreviation
        item_id (int, optional): the item class ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/item/{item_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def item_media(region_tag: str, item_id: int, *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an item by ID.

    Args:
        region_tag (str): region_tag abbreviation
        item_id (int): the creature family ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/item/{item_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def item_search(region_tag: str, field_values: Dict[str, Any], *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of items. For more detail see the search guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): search criteria as key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
         tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/item"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    #  adding locale and namespace key/values pairs to field_values to make a complete params list
    params.update(field_values)

    return uri, params


@verify_region
def item_appearance(region_tag: str, appearance_id: int, *, release: Optional[str] = None,
                    locale: Optional[str] = None) -> Tuple:
    """Returns an item appearance by ID.

    Args:
        region_tag (str): region_tag abbreviation
        appearance_id (int): the appearance ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
         tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/item-appearance/{appearance_id}"

    params = {
        locale: utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def item_appearance_search(region_tag: str, field_values: Dict[str, Any], *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of item appearances. For more detail see the search guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): search criteria as key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
         tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/item-appearance"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    #  adding locale and namespace key/values pairs to field_values to make a complete params list
    params.update(field_values)

    return uri, params


@verify_region
def item_appearance_set(region_tag: str, appearance_id: Optional[Union[int, str]] = "index", *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of appearances, or an appearance by ID

    Args:
        region_tag (str): region_tag abbreviation
        appearance_id (int, optional): the appearance ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/item-appearance/{appearance_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    return uri, params


@verify_region
def item_appearance_slot(region_tag: str, slot_type: Optional[str] = "index", *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of appearances, or an appearance by ID

    Args:
        region_tag (str): region_tag abbreviation
        slot_type: (str, optional): the appearance ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    if slot_type.upper() in ["HEAD", "NECK", "SHOULDER", "BODY", "CHEST", "WAIST", "LEGS", "FEET", "WRIST",
                             "HAND", "RING", "WEAPON", "SHIELD", "RANGED", "CLOAK", "TWOHWEAPON", "TABARD",
                             "ROBE", "WEAPONMAINHAND", "WEAPONOFFHAND", "HOLDABLE", "AMMO", "RANGEDRIGHT",
                             "PROFESSION_TOOL", "PROFESSION_GEAR", "EQUIPABLESPELL_WEAPON"]:
        uri = f"{utils.api_host(region_tag)}/data/wow/item-appearance/{slot_type.upper()}"
    else:
        uri = f"{utils.api_host(region_tag)}/data/wow/item-appearance/index"

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    return uri, params


@verify_region
def journal_expansion(region_tag: str, expansion_id: Optional[Union[int, str]] = "index", *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of journal expansions, or a journal expansion by ID.

    Args:
        region_tag (str): region_tag abbreviation
        expansion_id (int, optional): the encounter ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/journal-expansion/{expansion_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def journal_encounter(region_tag: str, encounter_id: Optional[Union[int, str]] = 'index', *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of journal encounters, or a journal encounter by ID.

    Notes:
        This replaced the Boss endpoint of the community REST API

    Args:
        region_tag (str): region_tag abbreviation
        encounter_id (int, optional): the encounter ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/journal-encounter/{encounter_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def journal_encounter_search(region_tag: str, field_values: Dict[str, Any], *,
                             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of journal encounters.  For more detail see the Search Guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): search criteria, as key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/journal-encounter"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    params.update(field_values)
    return uri, params


@verify_region
def journal_instance(region_tag: str, instance_id: Optional[Union[int, str]] = "index", *,
                     release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of journal instances, or a journal instance.

    Args:
        region_tag (str): region_tag abbreviation
        instance_id (int, optional): the encounter ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/journal-instance/{instance_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def journal_instance_media(region_tag: str, instance_id: int, *,
                           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a journal instance by ID.

    Args:
        region_tag (str): region_tag abbreviation
        instance_id (int): the creature family ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/journal-instance/{instance_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def media_search(region_tag: str, category: str,  field_values: Dict[str, Any], *,
                 release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of all types of media documents. For more detail see the Search Guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        category (str): media category
        field_values (dict): fields and values for the search criteria
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/media"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    #  adding locale and namespace key/values pairs to field_values to make a complete params list
    params.update({'tags': category})
    params.update(field_values)

    return uri, params


@verify_region
def modified_crafting(region_tag: str, release: Optional[str] = "retail", *,
                      locale: Optional[str] = 'enus') -> Tuple:
    """Returns the parent index for Modified Crafting.

    Args:
        region_tag (str): region_tag abbreviation
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/modified-crafting"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def modified_crafting_category(region_tag: str, category_id: Optional[Union[int, str]] = "index", *,
                               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the index of Modified Crafting categories, or a Modified Crafting category by ID.

    Args:
        region_tag (str): region_tag abbreviation
        category_id (int, optional): the encounter ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/modified-crafting/category/{category_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def modified_crafting_reagent_slot_type(region_tag: str, slot_type_id: Optional[Union[int, str]] = 'index', *,
                                        release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the index of Modified Crafting reagent slot types, or a Modified Crafting reagent slot type by ID

    Args:
        region_tag (str): region_tag abbreviation
        slot_type_id (int, optional): the encounter ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/modified-crafting/reagent-slot-type/{slot_type_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def mount(region_tag: str, mount_id: Optional[Union[int, str]] = "index", *,
          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of mounts, or a mount by ID.

    Args:
        region_tag (str): region_tag abbreviation
        mount_id (int, optional): the mount ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/mount/{mount_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def mount_search(region_tag: str, field_values: Dict[str, Any], *,
                 release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of mounts. For more detail see the Search Guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): fields and values for the search criteria
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/mount"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    #  adding locale and namespace key/values pairs to field_values to make a complete params list
    params.update(field_values)
    return uri, params


@verify_region
def mythic_keystone_affix(region_tag: str, affix_id: Optional[Union[int, str]] = "index", *,
                          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of mythic keystone affixes. or a mythic keystone affix by ID

    Args:
        region_tag (str): region_tag abbreviation
        affix_id (int, optional): the affix's ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/keystone-affix/{affix_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def mythic_keystone_affix_media(region_tag: str, affix_id: int, *,
                                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a mythic keystone affix by ID.

    Args:
        region_tag (str): region_tag abbreviation
        affix_id (int): the affix's ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/keystone-affix/{affix_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def mythic_keystone_dungeon(region_tag: str, dungeon_id: Optional[Union[int, str]] = "index", *,
                            release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of Mythic Keystone dungeons, or a Mythic Keystone dungeon by ID.

    Args:
        region_tag (str): region_tag abbreviation
        dungeon_id (int, optional): the dungeon's ID or 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/dungeon/{dungeon_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def mythic_keystone_index(region_tag: str, *,
                          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of links to other documents related to Mythic Keystone dungeons.

    Args:
        region_tag (str): region_tag abbreviation
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/index"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def mythic_keystone_period(region_tag: str, period_id: Optional[Union[int, str]] = "index", *,
                           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of Mythic Keystone periods, or a Mythic Keystone period by ID.

    Args:
        region_tag (str): region_tag abbreviation
        period_id (int, optional): the keystone's period ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = (
        f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/period/{period_id}"
    )
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def mythic_keystone_season(region_tag: str, season_id: Optional[Union[int, str]] = "index", *,
                           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of Mythic Keystone seasons, or a Mythic Keystone season by ID.

    Args:
        region_tag (str): region_tag abbreviation
        season_id (int, optional): the keystone's season ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = (
        f"{utils.api_host(region_tag)}/data/wow/mythic-keystone/season/{season_id}"
    )
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def mythic_keystone_leaderboard(region_tag: str, connected_realm_id: int, dungeon_id: Optional[Union[int, str]] = None,
                                period_id: Optional[Union[int, str]] = None, *,
                                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of Mythic Keystone Leaderboard dungeon instances for a connected realm,
    or a weekly Mythic Keystone Leaderboard by period.

    Args:
        release (str): release of the game (ie classic1x, classic, retail)
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        connected_realm_id (int): the connected realm's id
        dungeon_id (int, optional): the particular dungeon's ID or the word 'index'
        period_id (int, optional): the particular period to search or None when looking for the index

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/connected-realm/{connected_realm_id}/mythic-leaderboard/"

    if dungeon_id and period_id:
        uri += f"{dungeon_id}/period/{period_id}"
    else:
        uri += "index"

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def mythic_raid_leaderboard(region_tag: str, raid_name: str, faction: str, *,
                            release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the leaderboard for a given raid and faction.

    Args:
        region_tag (str): region_tag abbreviation
        raid_name (str): name of the raid
        faction (str): horde or alliance, defaults to alliance
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/leaderboard/hall-of-fame/"
    uri += f"{utils.slugify(raid_name)}/{utils.slugify(faction)}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def pet(region_tag: str, pet_id: Optional[Union[int, str]] = "index", *,
        release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of battle pets, or a battle pet by ID.

    Args:
        release (str): release of the game (ie classic1x, classic, retail)
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        pet_id (int, optional): the pet ID or the word 'index'

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pet/{pet_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pet_media(region_tag: str, pet_id: int, *,
              release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a battle pet by ID.

    Args:
        region_tag (str): region_tag abbreviation
        pet_id (int): the azerite pet ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/pet/{pet_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pet_ability(region_tag: str, pet_ability_id: Optional[Union[int, str]] = "index", *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of pet abilities, or a pet ability by ID.

    Args:
        region_tag (str): region_tag abbreviation
        pet_ability_id (int, optional): the pet ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pet-ability/{pet_ability_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pet_ability_media(region_tag: str, ability_id: int, *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a pet ability by ID.

    Args:
        region_tag (str): region_tag abbreviation
        ability_id (int): the azerite ability ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/pet-ability/{ability_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def playable_class(region_tag: str, class_id: Optional[Union[int, str]] = "index", *,
                   release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of playable classes, or a playable class by ID.

    Args:
        region_tag (str): region_tag abbreviation
        class_id (int, optional): the class ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/playable-class/{class_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def playable_class_media(region_tag: str, class_id: int, *,
                         release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a playable class by ID.

    Args:
        region_tag (str): region_tag abbreviation
        class_id (int ): class id
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/playable-class/{class_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pvp_talent_slots(region_tag: str, class_id: int, *,
                     release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the PvP talent slots for a playable class by ID.

    Args:
        region_tag (str): region_tag abbreviation
        class_id (int): class id
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/playable-class/{class_id}/pvp-talent-slots"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def playable_race(region_tag: str, race_id: Optional[Union[int, str]] = "index", *,
                  release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of playable races, or a playable race by ID.

    Args:
        region_tag (str): region_tag abbreviation
        race_id (int, optional): the playable race's ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/playable-race/{race_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def playable_spec(region_tag: str, spec_id: Optional[Union[int, str]] = "index", *,
                  release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    """Returns an index of playable specializations, or a playable specialization by ID.

    Args:
        region_tag (str): region_tag abbreviation
        spec_id (int, optional): the playable specialization's ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/playable-specialization/{spec_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def playable_spec_media(region_tag: str, spec_id: int, *,
                        release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a playable specialization by ID.

    Args:
        region_tag (str): region_tag abbreviation
        spec_id (int): specialization id
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/playable-specialization/{spec_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def power_type(region_tag: str, power_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of power types, or a power type by ID.

    Args:
        region_tag (str): region_tag abbreviation
        power_id (int, optional): the power type's ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/power-type/{power_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def profession(region_tag: str, profession_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """RReturns an index of professions, or a profession by ID.

    Args:
        region_tag (str): region_tag abbreviation
        profession_id (int, optional): the profession ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/profession/{profession_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def profession_media(region_tag: str, profession_id: int, *,
                     release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a profession by ID.

    Args:
        region_tag (str): region_tag abbreviation
        profession_id (str):  profession ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/profession/{profession_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def profession_skill_tier(region_tag: str, profession_id: int, skill_tier_id: int, *,
                          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns a skill tier for a profession by ID.

    Args:
        region_tag (str): region_tag abbreviation
        profession_id (int): the profession ID
        skill_tier_id (int): the skill tier ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/profession/{profession_id}/skill-tier/{skill_tier_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def recipe(region_tag: str, recipe_id: int, *,
           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns a recipe by ID.

    Args:
        region_tag (str): region_tag abbreviation
        recipe_id (str): the recipe ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/recipe/{recipe_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def recipe_media(region_tag: str, recipe_id: int, *,
                 release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a recipe by ID.

    Args:
        region_tag (str): region_tag abbreviation
        recipe_id (int): the profession ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/recipe/{recipe_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pvp_season(region_tag: str, season_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of PvP seasons, or a PvP season by ID.

    Args:
        region_tag (str): region_tag abbreviation
        season_id (int, optional): the power type's ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-season/{season_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def pvp_regions(region_tag: str, *, release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of PvP regions

    Args:
        region_tag (str): region_tag abbreviation
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-region/index"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def pvp_regional_season(region_tag: str, pvp_region_id: int, pvp_season_id: Optional[Union[int, str]] = 'index', *,
                        release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of PvP regions

    Args:
        region_tag (str): region_tag abbreviation
        pvp_region_id (int): the regional PVP ID (use pvp_regions)
        pvp_season_id (int): the pvp season ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-region/{pvp_region_id}/pvp-season/{pvp_season_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def pvp_leaderboard(region_tag: str, pvp_region_id: int, season_id: int, pvp_bracket: Optional[str] = "index", *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of PvP leaderboards for a PvP season, or
    the PvP leaderboard of a specific PvP bracket for a PvP season.

    Args:
        region_tag (str): region_tag abbreviation
        pvp_region_id (int): the regional PVP ID (use pvp_regions)
        season_id (int): pvp season's ID
        pvp_bracket (int, optional): the PvP bracket to view ('2v2', '3v3', '5v5', 'rbg') or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-region/{pvp_region_id}/pvp-season/"
    uri += f"{season_id}/pvp-leaderboard/{pvp_bracket}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def pvp_rewards_index(region_tag: str, pvp_region_id: int, season_id: int, *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of PvP rewards for a PvP season.

    Args:
        region_tag (str): region_tag abbreviation
        pvp_region_id (int): the regional PVP ID (use pvp_regions)
        season_id (int): the season ID for the rewards or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-region/{pvp_region_id}/pvp-season/{season_id}/pvp-reward/index"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def pvp_tier(region_tag: str, tier_id: Optional[Union[int, str]] = "index", *,
             release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of PvP tiers, or a PvP tier by ID.

    Args:
        region_tag (str): region_tag abbreviation
        tier_id (int, optional): the pvp tier ID or the default 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-tier/{tier_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pvp_tier_media(region_tag: str, tier_id: int, *,
                   release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a PvP tier by ID.

    Args:
        region_tag (str): region_tag abbreviation
        tier_id (int): pvp tier id
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/pvp-tier/{tier_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def quest(region_tag: str, quest_id: Optional[Union[int, str]] = "index", *,
          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the parent index for quests, or a quest by ID.

    Args:
        region_tag (str): region_tag abbreviation
        quest_id (int, optional): the quest ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/quest/{quest_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def quest_category(region_tag: str, quest_category_id: Optional[Union[int, str]] = "index", *,
                   release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of quest categories (such as quests for a specific class, profession, or storyline),
    or a quest category by ID.

    Args:
        region_tag (str): region_tag abbreviation
        quest_category_id (int, optional): the quest category ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = (
        f"{utils.api_host(region_tag)}/data/wow/quest/category/{quest_category_id}"
    )
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def quest_area(region_tag: str, quest_area_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of quest areas, or a quest area by ID.

    Args:
        region_tag (str): region_tag abbreviation
        quest_area_id (int, optional): the quest area ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/quest/area/{quest_area_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def quest_type(region_tag: str, quest_type_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of quest types (such as PvP quests, raid quests, or account quests),
    or a quest type by ID.

    Args:
        release (str): release of the game (ie classic1x, classic, retail)
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        quest_type_id (int, optional): the quest type ID or the word 'index'

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/quest/type/{quest_type_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def realm(region_tag: str, realm_slug: Optional[Union[str, int]] = "index", *,
          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of realms, or a single realm by slug or ID.

    Args:
        region_tag (str): region_tag abbreviation
        realm_slug (str/int, optional): the pvp tier ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/realm/{utils.slugify(realm_slug)}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def realm_search(region_tag: str, field_values: Dict[str, Any], *,
                 release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    """Performs a search of realms. For more detail see the Search Guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): search criteria, as key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/realm"
    params = {
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }
    params.update(field_values)
    return uri, params


@verify_region
def region(region_tag: str, region_req: Optional[Union[str, int]] = "index", *,
           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of regions, or a region by ID.

    Args:
        region_tag (str): region_tag abbreviation
        region_req (int, optional): the region_tag ID or the word 'index'
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/region/{region_req}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params


@verify_region
def reputation_faction(region_tag: str, faction_id: Optional[Union[int, str]] = "index", *,
                       release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    """Returns an index of reputation factions, or a single reputation faction by ID.

    Args:
        region_tag (str): region_tag abbreviation
        faction_id (int, optional): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/reputation-faction/{faction_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def reputation_tier(region_tag: str, tier_id: Optional[Union[int, str]] = "index", *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of reputation tiers, or a single set of reputation tiers by ID.

    Args:
        region_tag (str): region_tag abbreviation
        tier_id (int, optional): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/reputation-tiers/{tier_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def spell(region_tag: str, spell_id: int, *, release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns a spell by ID.

    Args:
        region_tag (str): region_tag abbreviation
        spell_id (int): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/spell/{spell_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def spell_media(region_tag: str, spell_id: int, *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a spell by ID.

    Args:
        release (str): release of the game (ie classic1x, classic, retail)
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        spell_id (int): pvp tier id

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/spell/{spell_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def spell_search(region_tag: str, field_values: Dict[str, Any] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Performs a search of spells. For more detail see the Search Guide.
    https://develop.battle.net/documentation/world-of-warcraft/guides/search

    Args:
        region_tag (str): region_tag abbreviation
        field_values (dict): search criteria, as key/value pairs
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/search/spell"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    #  adding locale and namespace key/values pairs to field_values to make a complete params list
    params.update(field_values)
    return uri, params


@verify_region
def talent_tree(region_tag: str, tree_id: Optional[int] = 'index', spec_id: Optional[int] = None,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the URL and parameters requried by the talent tree API

    Args:
        region_tag (str): region_tag abbreviation
        tree_id (int, optional): the tree ID
        spec_id (int, optional): the specialization ID
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/talent-tree/{tree_id}"

    if spec_id:
        uri = f"{uri}/playable-specialization/{spec_id}"

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def talent(region_tag: str, talent_id: Optional[Union[int, str]] = "index", *,
           release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of talents, or a talent by ID.

    Args:
        region_tag (str): region_tag abbreviation
        talent_id (int, optional): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/talent/{talent_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def pvp_talent(region_tag: str, pvp_talent_id: Optional[Union[int, str]] = "index", *,
               release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    """Returns an index of PvP talents, or a PvP talent by ID.

    Args:
        region_tag (str): region_tag abbreviation
        pvp_talent_id (int, optional): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/pvp-talent/{pvp_talent_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def tech_talent_tree(region_tag: str, tree_id: Optional[Union[int, str]] = "index", *,
                     release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of tech talent trees, or a tech talent tree by ID.

    Args:
        region_tag (str): region_tag abbreviation
        tree_id (int, optional): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/tech-talent-tree/{tree_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def tech_talent(region_tag: str, talent_id: Optional[Union[int, str]] = "index", *,
                release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns an index of tech talents, or a tech talent by ID.

    Args:
        region_tag (str): region_tag abbreviation
        talent_id (int, optional): the slug or ID of the region_tag requested
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/tech-talent/{talent_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def tech_talent_media(region_tag: str, talent_id: int, *,
                      release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns media for a tech talent by ID.

    Args:
        region_tag (str): region_tag abbreviation
        talent_id (int): pvp tier id
        release (str): release of the game (ie classic1x, classic, retail)
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/media/tech-talent/{talent_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def title(region_tag: str, title_id: Optional[Union[int, str]] = "index", *,
          release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    # noinspection GrazieInspection
    """Returns an index of titles, or a title by ID.

        Args:
            region_tag (str): region_tag abbreviation
            title_id (int, optional): the slug or ID of the region_tag requested
            release (str): release of the game (ie classic1x, classic, retail)
            locale (str): which locale to use for the request

        Returns:
            tuple: The URL (str) and parameters (dict)
        """
    uri = f"{utils.api_host(region_tag)}/data/wow/title/{title_id}"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }

    return uri, params


@verify_region
def toy(region_tag: str, toy_id: Optional[Union[str, int]] = "index", *,
        release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:

    uri = f"{utils.api_host(region_tag)}/data/wow/toy/{toy_id}"

    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "static", release)
    }
    return uri, params


@verify_region
def wow_token_index(region_tag: str, *,
                    release: Optional[str] = None, locale: Optional[str] = None) -> Tuple:
    """Returns the WoW Token index.

    Args:
        release (str): release of the game (ie classic1x, classic, retail)
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{utils.api_host(region_tag)}/data/wow/token/index"
    params = {
        "locale": utils.localize(locale),
        "namespace": wow_utils.namespace(region_tag, "dynamic", release)
    }

    return uri, params



"""Defines the Region for Diablo III

Classes:
    Region

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from battlenet_client.constants import Region as BaseRegion


__version__ = '2.0.0'
__author__ = 'David \'Gahd\' Couples'


class Region(BaseRegion):
    """Defines the constants for regions"""
    pass



"""Defines the Achievement related classes for WoW and WoW Classic

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.online>
"""
import json

from battlenet_client.wow.base import *
from battlenet_client.wow.game_data import quest, quest_category, quest_type, quest_area


class Quests(WoWBases):
    """Index of the Quests

    Args:
        region_tag (str): the region for the client

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): the localization string
    """
    
    quests: List[Quest] = field(init=False)

    def __init__(self):

        (url, params) = quest(region_tag, quest_id, release=release, locale=locale)
        if self._cache.check(self.locale, self.release, 'quest'):
            results = self._cache.select(self.locale, self.release, 'quest')
            init_data = results['data']
        else:
            init_data = client.get(url, params=params)
            if init_data.status_code == 404:
                raise BNetValueError("Invalid object ID")
            
            self._cache.submit(self.locale, self.release, 'quest', init_data.content)
                
        api_data = json.loads(init_data)
                              
        self.quests = api_data['quests']
       

class Quest(WoWBase):
    """Index of the Quests

    Args:
        region_tag (str): the region for the client

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): the localization string
    """
    title: str = field(init=False)
    required_level: int = field(init=False)
    area: QuestArea = field(init=False)
    description: str = field(init=False)
    recommended_levels: tuple = field(init=False)
    requirements: List = field(init=False)
    rewards: List = field(init=False)

    def __init__(self, client: Union[OAuth2Session, OicClient], region_tag: str, quest_id: Union[str, int], *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = quest(region_tag, quest_id, release=release, locale=locale)
        
        if self._cache.check(self.locale, self.release, 'quest]', self.object_id):
            results = self._cache.select(self.locale, self.release, 'quest', self.object_id)
            init_data = results['data']
        else:
            init_data = client.get(url, params=params)
            if init_data.status_code == 404:
                raise BNetValueError("Invalid object ID")
            
            self._cache.submit(self.locale, self.release, 'quest', init_data, self.object_id.content)

        api_data = json.loads(init_data)

        self.object_id = api_data['id']
        self.title = api_data['title']
        self.required_level = int(api_data['reqLevel'])
        self.area = QuestArea(client, region_tag, api_data['area']['id'], release=release, locale=locale)
        self.description = api_data['description']
        self.recommended_levels = (int(api_data['recommended_minimum_level']),
                                   int(api_data['recommended_maximum_level']))
        self.requirements = api_data['requirements']
        self.rewards = api_data['rewards']


class QuestCategory:
    """Index of Quest Categories

     Args:
         region_tag (str): region desired for the request\

     Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        quest_categories (list): the list of the available quest categories
    """
    def __init__(self,  client: Union[OAuth2Session, OicClient], region_tag: str, category_id: Union[str, int], *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = quest_category(region_tag, category_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not category_id:
            self.quest_categories = data['quest_categories']
        else:
            self.object_id = api_data['id']
            self.name = api_data['category']
            self.quests = api_data['quests']


class QuestArea:
    """Index of Quest Areas

    Args:
        region_tag (str): region desired for the request\

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        quest_areas (list): the list of the quest areas
    """
    def __init__(self,  client: Union[OAuth2Session, OicClient], region_tag: str, area_id: Union[str, int], *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = quest_area(region_tag, area_id, release=release, locale=locale)
        if self._cache.check(self.locale, self.release, 'playable_race]', self.object_id):
            results = self._cache.select(self.locale, self.release, 'playable_race', self.object_id)
            init_data = results['data']
        else:

            init_data = client.get(url, params=params)
            if init_data.status_code == 404:
                raise BNetValueError("Invalid object ID")
            
            self._cache.submit(self.locale, self.release, 'playable_race', init_data, self.object_id.content)

        api_data = json.loads(init_data)
        if not area_id:
            self.quest_areas = data['quests']
        else:
            self.area_id = api_data['id']
            self.name = api_data['name']
            self.quests = api_data['quests']


class QuestType:
    """Index of Quest Types

    Args:
        region_tag (str or :obj:`Region`): region desired for the request\

    Keyword Args:
         release (str, optional): the release to use.
         locale (str, optional): the localization to use from the API, The default of None means all localizations
         scope (list of str, optional): the scope or scopes to use during the endpoints that require the
             Web Application Flow
         redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
         client_id (str): the client ID from the developer portal
         client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): localization
        quest_types (list): the list of the quest areas
    """
    def __init__(self,  client: Union[OAuth2Session, OicClient], region_tag: str, type_id: Union[str, int], *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = quest_type(region_tag, type_id, release=release, locale=locale)

        api_data = client.get(url, params=params)

        if not type_id:
            self.quest_types = data['types']
        else:
            self.type_id = api_data['id']
            self.name = api_data['type']
            self.quests = api_data['quests']



from battlenet_client.cache.cache import Cache
from battlenet_client.utils import localize, next_update


import sqlite3
import datetime
from typing import Optional
import pathlib
import hashlib
from requests import Response
from abc import ABC, abstractmethod


__author__ = 'David "Gahd" Couples'
__version__ = "1.0.0"


class SQLite(Cache):
    """SQLite backend for cache

    Attributes:
        connection: connection of the database
        cursor: cursor of the database

    """

    @staticmethod
    def adapt_datetime_iso(val):
        """Adapt datetime.datetime to timezone-naive ISO 8601 date."""
        return val.isoformat()

    @staticmethod
    def convert_datetime(val):
        return datetime.datetime.fromisoformat(val.decode())

    def __init__(self, game: str, *, reset: Optional[bool] = False):
        """ Initializes access to the database

        Args:
            filename (str): the filename for the sqlite database

        Keywords:
            reset (bool): flag to reset the database (drop then create empty tables)

        """
        super().__init__(game)

        sqlite3.register_adapter(datetime.datetime, SQLite.adapt_datetime_iso)
        sqlite3.register_converter("datetime", SQLite.convert_datetime)
        self.connection = sqlite3.connect(f"{self.game}_cache.db", 
                                          detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
        self.connection.row_factory = sqlite3.Row
        self.cursor = self.connection.cursor()

        if reset:
            current_dir = pathlib.Path(__file__)
            with open(current_dir / 'cache_db.sql', "r") as config:
                self.cursor.executescript(config.read())

    def __str__(self):
        return f"{self.__class__.__name__} {__version__}"

    def __repr__(self):
        return f"{self.__class__.__name__} {c__version__}"

    def select(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> bytes:
        """Gathers data from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bytes): Returns the data from the cache
        """
        hash_value = hashlib.sha1(f"{game}:{region}:{category}:{release}:{obj_id}".encode()).hexdigest()
        return self.cursor.execute('''select data from cache where hash = ?;''', (hash_value,
                                                                                  )).fetchone()['data']

    def select_all(self, region: str, release: str, category: str, obj_id: Optional[int] = None):
        """Gathers data from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bytes): Returns the data from the cache
        """
        hash_value = hashlib.sha1(f"{game}:{region}:{category}:{release}:{obj_id}".encode()).hexdigest()
        results = self.cursor.execute('''select data from cache where hash = ?;''', (hash_value,
                                                                                  )).fetchall()
        return [result['data'] for result in results]

    def delete(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> int:
        """Removes a row from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (int): number of rows deleted
        """
        hash_value = hashlib.sha1(f"{game}:{region}:{category}:{release}:{obj_id}".encode()).hexdigest()
        self.cursor.execute('''delete from cache
         where hash = ?;''', (hash_value,))

        self.connection.commit()
        return self.cursor.rowcount

    def submit(self, region: str, release: str, category: str, response: Response,
               obj_id: Optional[int] = None) -> int:
        """Gathers data from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the data's category
            response (:obj:`Response`): the response from the API
            obj_id (int, optional): the ID for the cached data

        Returns:
            (int): number of rows inserted or updated
        """
        hash_value = hashlib.sha1(f"{game}:{region}:{category}:{release}:{obj_id}".encode()).hexdigest()

        self.cursor.execute('''insert into cache (hash, category, data, next_update) 
                select ?, ?, ?, ?                
                on conflict (hash) do update
                  set data = excluded.data, next_update = excluded.next_update 
                  where next_update < datetime('now');''', (hash_value, cat['ROWID'], response.content,
                                                            self.next_update(response.headers['last-modified'],
                                                                             category)))
        self.connection.commit()
        return self.cursor.rowcount

    def check(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> bool:
        """Returns if the data is found and if it is expired

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the data's category
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bool): True if the object is found and has not expired. False otherwise
        """
        hash_value = hashlib.sha1(f"{game}:{region}:{category}:{release}:{obj_id}".encode()).hexdigest()
        results = self.cursor.execute('''select datetime('now') > next_update as expired from cache
        where hash = ?;''', (hash_value,)).fetchone()

        if not results or results['expired']:
            return False

        return True

    def category(self, name: str):
        """Returns the category data from the cache

        Args:
            name (str): name of the category

        Returns:
            (:obj:`sqlite3.Row`): the row of for the provided category
        """
        return self.cursor.execute('''select * from categories where name = ?''',
                                   (name,)).fetchone()




"""Defines the functions that handle the community APIs for Diablo III

Functions:
    act(region_tag, act_id, locale)
    artisan(region_tag, artisan_slug, locale)
    recipe(region_tag, artisan_slug, recipe_slug, locale)
    follower(region_tag, follower_slug, locale)
    character_class(region_tag, class_slug, locale)
    api_skill(region_tag, class_slug, skill_slug, locale)
    item_type(region_tag, item_type_slug, locale)
    item(region_tag, item_slug, locale)
    api_account(region_tag, bnet_tag, locale)
    api_hero(region_tag, bnet_tag, hero_id, category, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from typing import Union, Optional
from urllib.parse import quote as urlquote

from ..decorators import verify_region
from ..exceptions import BNetValueError
from ..utils import slugify, localize, api_host


__version__ = "2.0.0"
__author__ = "David \"Gahd\" Couples"


@verify_region
def act(
    region_tag: str,
    *,
    act_id: Optional[Union[int, str]] = None,
    locale: Optional[str] = None
):
    """Returns the index of acts, or the act by ID

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        act_id (int, optional): the act's ID to retrieve its data

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/data/act"

    if act_id:
        uri += f"/{act_id}"

    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def artisan(
    region_tag: str, artisan_slug: str, locale: Optional[str] = None
):
    """Returns a single artisan by the slug

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        artisan_slug (str): the slug of the artisan

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/data/artisan/{slugify(artisan_slug)}"
    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def recipe(
    region_tag: str,
    artisan_slug: str,
    recipe_slug: str,
    locale: Optional[str] = None,
):
    """Returns a single recipe by slug for the specified artisan.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        artisan_slug (str): the slug of the artisan
        recipe_slug (str): the slug of the recipe

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/data/artisan/{slugify(artisan_slug)}/recipe"
    uri += f"/{slugify(recipe_slug)}"
    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def follower(
    region_tag: str, follower_slug: str, locale: Optional[str] = None
):
    """Returns a single follower by slug.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        follower_slug (str): the slug of a follower

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/data/follower/{slugify(follower_slug)}"
    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def character_class(
    region_tag: str, class_slug: str, locale: Optional[str] = None
):
    """Returns a single character class by slug.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        class_slug (str): the slug of a character class

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/data/hero/{slugify(class_slug)}"
    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def api_skill(
    region_tag: str,
    class_slug: str,
    skill_slug: str,
    locale: Optional[str] = None,
):
    """Returns a single skill by slug for a specific character class.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        class_slug (str): the slug of a character class
        skill_slug (str):

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/d3/data/hero/{slugify(class_slug)}"
    uri += f"/skill/{slugify(skill_slug)}"
    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def item_type(
    region_tag: str,
    *,
    item_type_slug: Optional[str] = None,
    locale: Optional[str] = None,
):
    """Returns an index of item types, or a single item type by slug

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        item_type_slug (str, optional): the slug of an item type

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/d3/data/item-type"

    if item_type_slug:
        uri += f"/{slugify(item_type_slug)}"

    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def item(region_tag: str, item_slug: str, locale: Optional[str] = None):
    """Returns a single item by item slug and ID.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        item_slug (str): the slug of the item

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/data/item/{item_slug}"
    print(uri)
    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def api_account(
    region_tag: str, bnet_tag: str, locale: Optional[str] = None
):
    """Returns the specified account profile.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        bnet_tag (str): bnet tag of the user

    Returns:
        tuple: The URL (str) and parameters (dict)
    """

    uri = f"{api_host(region_tag)}/d3/profile/{urlquote(bnet_tag)}/"

    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def api_hero(
    region_tag: str,
    bnet_tag: str,
    hero_id: str,
    *,
    category: Optional[str] = None,
    locale: Optional[str] = None
):
    """Returns a single hero, a list of items for the specified hero, or
    list of items for the specified hero's followers.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        bnet_tag (str): BNet tag for the account
        hero_id (str):  Hero's ID
        category (str): category to retrieve if specified ('items', 'follower-items')

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetValueError: category is not None and not 'items' or 'follower-items'
    """
    uri = f"{api_host(region_tag)}/d3/profile/{urlquote(bnet_tag)}/"
    uri += f'hero/{hero_id}'

    params = {"locale": localize(locale)}

    if category:

        if category not in ("items", "follower-items"):
            raise BNetValueError(
                "Invalid category;  Valid categories are 'items' and 'follower-items'"
            )

        uri += f"/{category.lower()}"

    return uri, params



"""Defines the decorators used in this package

Functions:
   verify_region

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
import functools

from .constants import VALID_REGIONS
from .exceptions import BNetRegionError


__version__ = '1.0.0'
__author__ = 'David \'Gahd\' Couples'


def verify_region(func):
    """Verifies the function's first parameter is a valid region abbreviation

    Raises:
        BNetRegionNotFoundError: when the region (arg[0]) is not a valid region tag
    """
    @functools.wraps(func)
    def wrapper_verify_region(*args, **kwargs):
        if args[0].lower() not in VALID_REGIONS:
            raise BNetRegionError('Region not found')

        return func(*args, **kwargs)
    return wrapper_verify_region




"""Defines exceptions related to the Battle.net OAuth API wrappers

Disclaimer:
    All rights reserved, Blizzard is the intellectual property owner of Battle.net and any data
    retrieved from this API.
"""


class BNetError(Exception):
    """Base Error class for BattleNet Client"""

    pass


class BNetRegionNotFoundError(BNetError):
    """Error raised when an invalid region bnet is detected"""

    pass


class BNetClientError(BNetError):
    """Error raised if there is a mismatch with the client and API endpoints"""

    pass



"""Defines the Achievement related classes for WoW and WoW Classic

This module contains the client class definitions for accessing World of Warcraft API endpoints.
There are two flavors of clients, one implements the client credential workflow, which happens
to be the most common.  The other implements the user authorization workflow

Examples:
    > from wow_api import WoWCredentialClient
    > client = WoWCredentialClient(<region>, <locale>, client_id='<client ID>', client_secret='<client secret>')
    > achievement_category = AchievementCategory(client, <locale>, category_id=3)

Defines the various game data API requests for World of Warcraft and
World of Warcraft Classic

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.io>
"""



# TODO: v4.2.0 add crafting




"""Defines the functions that handle the community APIs for Starcraft 2
for the chinese regions

Functions:
    profile(region_tag, profile_id, realm_id, profile_name, locale)
    ladder(region_tag, profile_id, realm_id, profile_name, locale)
    match_history( region_tag, profile_id, realm_id, profile_name, locale)
    ladder( region_tag, ladder_id, locale)
    achievements(region_tag, locale)
    rewards(region_tag, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from typing import Optional

from .. import utils
from ..exceptions import BNetRegionError
from ..decorators import verify_region
from .decorators import sc2_cn_only


__version__ = '2.0.0'
__author__ = 'David \'Gahd\' Couples'


@sc2_cn_only
@verify_region
def profile(
    region_tag: str,
    profile_id: int,
    region_id: int,
    profile_name: str,
    locale: Optional[str] = None
):
    """Returns data about an individual SC2 profile.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        region_id (int): region for the profile, or use sc2.constants
        profile_id (int): the profile ID
        profile_name (str): name to use with the CN endpoint

    Returns:
        tuple: The URL (str) and parameters (dict)

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetRegionError: when the region tag is not 'cn'
    """
    if region_tag != 'cn':
        raise BNetRegionError("CN Region required")

    uri = f"{utils.api_host(region_tag)}/sc2/profile/{profile_id}/{region_id}/{profile_name}"

    params = {"locale": utils.localize(locale)}
    return uri, params


@sc2_cn_only
@verify_region
def ladders(
    region_tag: str,
    profile_id: int,
    region_id: int,
    profile_name: str,
    locale: Optional[str] = None
):
    """Returns a ladder summary, or specific ladder for an individual SC2 profile.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        region_id (int): region for the profile, or use sc2.constants
        profile_id (int): the profile ID
        profile_name (str): Name of the profile

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetRegionError: when the region tag is not 'cn'
    """
    if region_tag != 'cn':
        raise BNetRegionError("CN Required")

    uri = f"{utils.api_host(region_tag)}/sc2/profile/{profile_id}/{region_id}/{profile_name}/ladders"

    params = {"locale": utils.localize(locale)}

    return uri, params


@sc2_cn_only
@verify_region
def match_history(
        region_tag: str,
        profile_id: int,
        region_id: int,
        profile_name: str,
        locale: Optional[str] = None
):
    """Returns data about an individual SC2 profile's match history.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        region_id (int): region for the profile, or use sc2.constants
        profile_id (int): the profile ID
        profile_name(str): profile name for the CN region request

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetRegionError: when the region is not 'cn'
    """
    uri = f"{utils.api_host(region_tag)}/sc2/profile/{profile_id}/{region_id}/{profile_name}/matches"

    params = {"locale": utils.localize(locale)}

    return uri, params


@sc2_cn_only
@verify_region
def ladder(
    region_tag: str,
    ladder_id: int,
    locale: Optional[str] = None,
):
    """Returns a ladder summary, or specific ladder for an individual SC2 profile.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        ladder_id (int): region for the profile, or use sc2.constants

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetRegionError: when the region tag is not 'cn'
    """
    if region_tag != 'cn':
        raise BNetRegionError("CN Required")

    uri = f"{utils.api_host(region_tag)}/sc2/ladder/{ladder_id}"

    params = {"locale": utils.localize(locale)}

    return uri, params


@sc2_cn_only
@verify_region
def achievements(
    region_tag: str,
    locale: Optional[str] = None,
):
    """Returns a ladder summary, or specific ladder for an individual SC2 profile.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetRegionError: when the region tag is not 'cn'
    """
    if region_tag != 'cn':
        raise BNetRegionError("CN Required")

    uri = f"{utils.api_host(region_tag)}/sc2/data/achievements"

    params = {"locale": utils.localize(locale)}

    return uri, params


@sc2_cn_only
@verify_region
def rewards(
    region_tag: str,
    locale: Optional[str] = None,
):
    """Returns a ladder summary, or specific ladder for an individual SC2 profile.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)

    Raises:
        BNetRegionError: when the region tag is not 'cn'
    """
    if region_tag != 'cn':
        raise BNetRegionError("CN Required")

    uri = f"{utils.api_host(region_tag)}/sc2/data/rewards"

    params = {"locale": utils.localize(locale)}

    return uri, params



"""Miscellaneous functions to support for Battle.net

Functions:
    currency_convertor(value)
    slugify(value)
    localize(locale)
    api_host(region_tag)
    auth_host(region_tag)
    render_host(region_tag)
    nth_occurrence(string, sep, count)
    r_nth_occurrence(string, sep, count)

Misc Variables:
    WOW_CLASSICS


Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from battlenet_client.exceptions import BNetValueError

from typing import Tuple, Optional, Union
from datetime import datetime, timedelta


__version__ = '1.0.0'
__author__ = 'David \'Gahd\' Couples'


def currency_convertor(value: int) -> Tuple[int, int, int]:
    """Returns the value into gold, silver and copper

    Args:
        value (int): the value to be converted

    Returns:
        tuple: gold (int), silver (int) and copper (int)
    """    
    if value < 0:
        raise BNetValueError("Value cannot be negative")

    return int(value) // 10000, (value % 10000) // 100, int(value) % 100
    

def slugify(value: Union[int, str]) -> str:
    """Returns value as a slug

    Args:
        value (str): the string to be converted into a slug

    Returns:
        str: the slug
    """
    if isinstance(value, int):
        return str(value)

    return value.lower().replace("'", "").replace(" ", "-")


def localize(locale: Optional[str] = None) -> Union[None, str]:
    """Returns the standardized locale

    Args:
        locale (str): the locality to be standardized

    Returns:
        str: the locale in the format of "<language>_<DIALECT>"

    Raise:
        TypeError: when locale is not a string
        ValueError: when the lang and country are not in the given lists
    """
    if not locale:
        return None

    if not isinstance(locale, str):
        raise BNetValueError("Locale must be a string")

    languages = ("en", "es", "pt", "fr", "ru", "de", "it", "ko", "zh")
    if locale[:2].lower() not in languages:
        raise BNetValueError("Invalid language")

    dialects = ("us", "mx", "br", "gb", "es", "fr", "ru", "de", "pt", "it", "kr", "tw", "cn")
    if locale[-2:].lower() not in dialects:
        raise BNetValueError("Invalid country")

    return f"{locale[:2].lower()}_{locale[-2:].upper()}"


def api_host(region_tag: str) -> str:
    """Returns the API endpoint hostname and protocol

    Args:
        region_tag (str): the region abbreviation

    Returns:
        str: The API endpoint hostname and protocol
    """
    if region_tag.lower() == "cn":
        return "https://gateway.battlenet.com.cn"

    return f"https://{region_tag.lower()}.api.blizzard.com"


def auth_host(region_tag: str) -> str:
    """Returns the authorization endpoint hostname and protocol

    Args:
        region_tag (str): the region abbreviation

    Returns:
        str: The authorization endpoint hostname and protocol
    """
    if region_tag.lower() == "cn":
        return "https://www.battlenet.com.cn"

    if region_tag.lower() in ("kr", "tw"):
        return "https://apac.battle.net"

    return f"https://{region_tag.lower()}.battle.net"


def render_host(region_tag: str) -> str:
    """Returns the render endpoint hostname and protocol

    Args:
        region_tag (str): the region abbreviation

    Returns:
        str: The render endpoint hostname and protocol
    """
    if region_tag.lower() == "cn":
        return "https://render.worldofwarcraft.com.cn"

    return f"https://render-{region_tag.lower()}.worldofwarcraft.com"


def nth_occurrence(string: str, sep: str, count: int) -> str:
    """Returns `path` ending at the `count` occurrence of `sep`

    Args:
        string (str): the original string
        sep (str): the separator that needs to be counted
        count (int): the number of occurrences of `sep`
    """
    start_index = string.find(sep)

    while start_index >= 0 and count > 1:
        start_index = string.find(sep, start_index + 1)
        count -= 1

    return string[:start_index]


def r_nth_occurrence(string: str, sep: str, count: int) -> str:
    """Returns `path` ending at the `count` occurrence of `sep`

    Args:
        string (str): the original string
        sep (str): the separator that needs to be counted
        count (int): the number of occurrences of `sep`
    """
    end_index = string.rfind(sep)

    while end_index >= 0 and count > 1:
        end_index = string.rfind(sep, 0, end_index - 1)
        count -= 1

    return string[end_index + 1:]


def next_update(last_modified: str, duration: float) -> datetime:
    """Returns the datetime object representing the last modified of the API endpoint

    Args:
        last_modified (str): 'last modified' from the response header
        duration (float): the duration in weeks for the data

    Returns:
          datetime: datetime of the next
    """
    last_modified = datetime.strptime(last_modified, '%a, %d %b %Y %H:%M:%S %Z')
    last_modified += timedelta(weeks=duration)
    return last_modified



"""Defines the functions that handle the game data APIs for Hearthstone

Functions:
    card_search(region_tag, field_values, locale)
    card(region_tag, card_id, game_mode, locale)
    metadata(region_tag, meta_data, locale)
    card_back_search(region_tag, field_values, locale)
    card_back(region_tag, card_back_id, locale)
    card_deck(region_tag, field_values, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""

from typing import Optional, Any, Dict

from .. import utils
from ..decorators import verify_region
from ..exceptions import BNetValueError


__version__ = '2.0.0'
__author__ = 'David \'Gahd\' Couples'


@verify_region
def card_search(
    region_tag: str,
    field_values: Dict[str, Any],
    locale: Optional[str] = None,
):
    """Returns an up-to-date list of all cards matching the search criteria.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        field_values (dict): search criteria, as key/value pairs
            For more information for the field names and options:
            https://develop.battle.net/documentation/hearthstone/game-data-apis

    Returns:
        dict: json decoded search results that match `field_values`

    """
    params = {"locale": utils.localize(locale)}

    if "gameMode" not in field_values.keys():
        params["gameMode"] = "constructed"

    uri = f"{utils.api_host(region_tag)}/hearthstone/cards"

    if field_values:
        params.update(field_values)

    return uri, params


@verify_region
def card(
    region_tag: str,
    card_id: str,
    *,
    game_mode: Optional[str] = "constructed",
    locale: Optional[str] = None,
):
    """Returns the card with an ID or slug that matches the one you specify.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        card_id (int, str): the ID or full slug of the card
        game_mode (str, optional): the game mode
            See for more information:
            https://develop.battle.net/documentation/hearthstone/guides/game-modes

    Returns:
        dict: json decoded data for the index/individual azerite essence(s)
    """
    uri = f"{utils.api_host(region_tag)}/hearthstone/cards/{card_id}"
    params = {"locale": utils.localize(locale), "gameMode": game_mode}

    return uri, params


@verify_region
def metadata(
    region_tag: str,
    *,
    meta_data: Optional[str] = None,
    locale: Optional[str] = None,
):
    """Returns information about the categorization of cards. Metadata
    includes the card set, set group (for example, Standard or Year of
    the Dragon), rarity, class, card type, minion type, and keywords, or
    information about just one type of metadata.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        meta_data (str, optional): what metadata to filter
            Please see below for more information
            https://develop.battle.net/documentation/hearthstone/guides/metadata
            valid options: `sets`, `setGroups`, `types`, `rarities`, `classes`,
            `minionTypes`, `keywords`

    Returns:
        dict: json decoded list of metadata or a specific set of metadata
    """
    categories = ('sets', 'setGroups', 'types', 'rarities', 'classes', 'minionTypes', 'keywords')

    if meta_data and meta_data not in categories:
        raise BNetValueError("Metadata category not valid")

    uri = f"{utils.api_host(region_tag)}/hearthstone/metadata"

    if meta_data:
        uri += f"/{meta_data}"

    params = {"locale": utils.localize(locale)}

    return uri, params


@verify_region
def card_back_search(
    region_tag: str,
    field_values: Dict[str, Any],
    locale: Optional[str] = None,
):
    """Returns an up-to-date list of all card backs matching the search criteria.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        field_values (dict): search criteria, as key/value pairs
            For more information for the field names and options:
            https://develop.battle.net/documentation/hearthstone/guides/card-backs

    Returns:
        dict: json decoded search results that match `field_values`
    """
    uri = f"{utils.api_host(region_tag)}/hearthstone/cardbacks"

    if 'textFilter' in field_values and not locale:
        raise BNetValueError('textFilter requires a locale')

    params = {"locale": utils.localize(locale)}

    if field_values:
        params.update(field_values)

    return uri, params


@verify_region
def card_back(
    region_tag: str, card_back_id: str, locale: Optional[str] = None
):
    """Returns a specific card back by using card back ID or slug.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        card_back_id (int, str): the ID or full slug of the card back

    Returns:
        dict: json decoded data for the card back identified by `card_back_id`
    """
    uri = f"{utils.api_host(region_tag)}/hearthstone/cardbacks/{utils.slugify(card_back_id)}"

    params = {"locale": utils.localize(locale)}

    return uri, params


@verify_region
def card_deck(
    region_tag: str,
    field_values: Dict[str, Any],
    locale: Optional[str] = None,
):
    """Finds a deck by list of cards, including the hero.

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request
        field_values (dict): search criteria, as key/value pairs
            For more information for the field names and options:
            https://develop.battle.net/documentation/hearthstone/guides/decks

    Returns:
        dict: json decoded search results that match `field_values`
    """
    uri = f"{utils.api_host(region_tag)}/hearthstone/deck"

    params = {"locale": utils.localize(locale)}

    if field_values:
        params.update(field_values)

    return uri, params



"""Defines the functions that handle the community APIs for Diablo III

Functions:
    summary(region_tag, locale)
    player(region_tag, player_id, locale)
    matches(region_tag, match_id, locale)
    follower(region_tag, follower_slug, locale)
    segment(region_tag, segment_id, locale)
    team(region_tag, team_id, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: September 13, 2023
"""

from typing import Union, Optional

from ..decorators import verify_region
from ..utils import localize, api_host


__version__ = "2.0.0"
__author__ = "David \"Gahd\" Couples"


@verify_region
def summary(
    region_tag: str,
    *,
    locale: Optional[str] = None
):
    """Returns a summary of league data

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/owl/v1/owl2"

    params = {"locale": localize(locale)}

    return uri, params


@verify_region
def player(
    region_tag: str,
    *,
    player_id: Union[str, int],
    locale: Optional[str] = None
):
    """Returns player data

    Args:
        region_tag (str): region_tag abbreviation
        player_id (str or int):  player ID to query
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/owl/v1/players/{player_id}"
    params = {"locale": localize(locale)}
    return uri, params


@verify_region
def match(
    region_tag: str,
    *,
    match_id: Union[str, int],
    locale: Optional[str] = None
):
    """Returns match data

    Args:
        region_tag (str): region_tag abbreviation
        match_id (str or int):  match ID to query
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/owl/v1/matches/{match_id}"
    params = {"locale": localize(locale)}
    return uri, params


@verify_region
def segment(
    region_tag: str,
    *,
    segment_id: Union[str, int],
    locale: Optional[str] = None
):
    """Returns match data

    Args:
       region_tag (str): region_tag abbreviation
       segment_id (str or int):  match ID to query
       locale (str): which locale to use for the request

    Returns:
       tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/owl/v1/segments/{segment_id}"
    params = {"locale": localize(locale)}
    return uri, params


@verify_region
def team(
    region_tag: str,
    *,
    team_id: Union[str, int],
    locale: Optional[str] = None
):
    """Returns team data

    Args:
        region_tag (str): region_tag abbreviation
        team_id (str or int):  match ID to query
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    uri = f"{api_host(region_tag)}/owl/v1/teams/{team_id}"
    params = {"locale": localize(locale)}
    return uri, params



"""Miscellaneous functions to support for World of Warcraft

Functions:
    namespace(api_type, release, region)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from ..decorators import verify_region
from ..exceptions import BNetValueError
from .constants import Release

from typing import Optional

__version__ = '1.0.0'
__author__ = 'David \'Gahd\' Couples'


@verify_region
def namespace(region: str, api_type: str, release: Optional[str] = 'retail') -> str:
    """Returns the namespace required by the WoW API endpoint

    Returns:
        str: the namespace string

    Raises:
        BNetValueError: when api type is not of static, dynamic or profile, or an invalid release
    """

    if api_type.lower() not in ('static', 'dynamic', 'profile'):
        raise BNetValueError('Invalid API type: needs to be static, dynamic, or profile')

    if not release:
        release = 'retail'

    if release.lower() not in Release.all():
        raise BNetValueError(f"Invalid Release: needs to be one of: {','.join(Release.all())}")

    if release.lower() != "retail":
        return f"{api_type}-{release.lower()}-{region.lower()}"

    return f"{api_type}-{region.lower()}"



"""Defines the Region for World of Warcraft

Classes:
    Region
    Release

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""
from ..constants import Region as BaseRegion


__version__ = '3.0.1'
__author__ = 'David \'Gahd\' Couples'

VALID_LOCALES = ('en_US', 'es_MX', 'pt_BR', 'ko_KR', 'en_GB', 'es_ES', 'fr_FR', 'ru_RU', 'de_DE', 'pt_PT', 'it_IT',
                 'zh_TW', 'zh_CN')


class Region(BaseRegion):
    class Id:
        """Defines the Region IDs for World of Warcraft"""

        #: Region ID for North America
        NORTH_AMERICA = 1

        #: Region ID for Taiwan
        TAIWAN = 4

        #: Region ID for Europe
        EUROPE = 3

        #: Region ID for Korea
        KOREA = 2

        #: Region ID for China
        CHINA = 5


class Release:
    """Defines the Release Names for World of Warcraft/World of Warcraft Classic"""

    #: Release name for the original World of Warcraft (v 1.0)
    VANILLA = "classic1x"

    #: Release name for re-release of Burning Crusade through latest classic expansion (v 2.0)
    CLASSIC = "classic"

    #: Release name for the current expansion
    RETAIL = "retail"

    @classmethod
    def all(cls) -> list:
        """Returns the list of all releases

        Returns:
            list: list of all releases
        """
        all_list = []
        for name in dir(cls):
            if name.startswith('__'):
                continue

            obj = getattr(cls, name)
            if isinstance(obj, str):
                all_list.append(obj)

        return all_list



"""World of Warcraft Mounts and related index

Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:


David "Gahd" Couples <gahdania@gahd.io>
"""

from battlenet_client.wow.game_data import mount


class Mount:
    """Defines the Playable Race Index.

    Args:
        region_tag (str): region abbreviation for use with the APIs

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        locale (str): the localization
        mounts (list): the index of the mounts
    """
    def __init__(self, client: Union[OAuth2Session, OicClient],  region_tag: str, mount_id: Optional[str, int] = None, *,
                 release=None, locale=None):

        (url, params) = mount(region_tag, mount_id, release=release, locale=locale)

        api_data = client.get(url, params=parms)

        if not mount_id:
            self.mounts = api_data['mounts']

        else:
            self.mount_id = api_data['id']
            self.name = api_data['name']
            self.creature_displays = api_data['creature_displays']
            self.description = api_data['description']
            self.source = api_data['source']['name']
            self.faction = api_data['faction']['name']
            self.requirements = api_data['requirements']



"""
Base class for the World of Warcraft classes

WoWBase
"""
from battlenet_client.cache.cache import Cache
from battlenet_client.utils import localize, r_nth_occurrence
from battlenet_client.exceptions import BNetValueError

import json
from typing import Optional, Union, Any, Dict
from requests_oauthlib import OAuth2Session
from requests import Response
from oic import oic
import functools
from pathlib import Path
from urllib import parse
from PIL import Image
from datetime import datetime, timedelta
from abc import ABCMeta, abstractmethod
import urllib.request

__version__ = "v1.0.0"


class WoWBase(object):
    """Base class for classes that handle specific aspects of the game.

    Attributes:
        region_tag (str):  lowercase alphabetical representation for the region the class belongs
        release (str): release of the game
        locale (str):  the localization to use for the requests
    """

    def __init__(self, region_tag: str, object_id: int, *, release: Optional[str] = 'retail',
                 locale: Optional[str] = None):
        self.id = object_id
        self.region_tag = region_tag.lower()
        self.release = release.lower()
        self.locale = localize(locale)

    def __repr__(self):
        return f"{self.__class__.__name__} {__version__}"

    def __str__(self):
        return f"{self.__class__.__name__} {__version__}"

    def __eq__(self, other):
        return self.id == other.id

    def get(self, client: Union[OAuth2Session, oic.Client], release: str, category: str, url: str,
            params: Dict[str, Any], *, object_id: Optional[int] = None,
            cache: Optional[Cache] = None) -> Dict[str, Any]:
        """Returns the data from either the cache or the API

        Args:
            client (:obj:`OAuth2Session` or :obj:`oic.Client`): the client to communicate with the API
            release (str): the release of World of Warcraft
            category (str): category the API request
            url (str): the URL to the API endpoint
            params (dict of str,str): parameters to pass to the API
            object_id (int, optional): the ID to pass to the API endpoint
            cache (:obj:`Cache`, optional): cache to use
        """

        if cache:
            if cache.check(self.region_tag, release, category, object_id):
                return json.loads(cache.select(self.region_tag, release, category, object_id))

        api_data = client.get(url, params=params)

        if api_data.status_code == 404:
            raise BNetValueError(api_data.text)

        if cache:
            cache.submit(self.region_tag, release, category, api_data, object_id)

        return json.loads(api_data.content)

    def _localized_data(self, field_name: str, locale: Optional[str] = None):
        """Returns the data from `field_name` localized by the given `locale`

        Args:
            field_name (str): the member variable to localize
            locale (str, optional): locale to select.

        Returns:
            (str): the localized value for `field_name`
        """

        attrib = getattr(self, field_name)
        locale = localize(locale or self.locale or 'en_US')

        if localization in attrib.keys():
            return attrib[localization]

        if self.locale in attrib.keys():
            return attrib[self.locale]

        return attrib['en_US']

    def _gender_data(self, field_name: str, gender: str, locale: Optional[str] = None):
        """Returns the selected form of `field_name` based on the provided `gender` and `locale`

        Args:
            field_name (str): the member variable to select
            locale (str, optional): the locale to select

        Returns:
            (str): the selected result from `field_name`
        """
        localized_data = getattr(self, field_name)[gender]
        localization = localize(locale or self.locale or 'en_US')
        return localized_data[localization]


class WoWMedia(object):
    """Media Storage

    Attributes:
        image (:obj:`PIL.Image`):  File pointer for image
        object_id (int): ID for the object
        assets (list of dict):  the assets associated with the object
    """
    def __init__(self, media_info: Dict[str, Any], fs_path: Optional[str] = '.'):
        """Initializes the Media object

        Args:
            media_info (dict): the data from the media API endpoint
            fs_path (str): the filesystem path to store the files
        """
        self._image = None
        self._blizzard_url = media_info['assets'][0]['value']
        self._file_id = media_info['assets'][0]['file_data_id']

        self.url = r_nth_occurrence(self._blizzard_url, '/', 3)
        self._file = Path(fs_path).parent / self.url

        if not self._file.parent.exists():
            self._file.parent.mkdir(parents=True)

        if not self._file.exists():
            urllib.request.urlretrieve(self._blizzard_url, self._file)

        mod_time = datetime.fromtimestamp(self._file.stat().st_mtime)
        if mod_time + timedelta(weeks=1) < datetime.now():
            self._file.unlink(missing_ok=True)
            urllib.request.urlretrieve(self._blizzard_url, self._file)

    def __del__(self):
        if self._image:
            self._image.close()

    def open(self):
        """Opens the media file
        """
        if self._file.exists():
            self._image = Image.open(self._file)

    def show(self):
        if not self._image:
            self.open()

        self._image.show()

    @property
    def size(self):
        if self._image:
            return self._image.size



"""Defines the Achievement related classes for WoW and WoW Classic


Disclaimer:
All rights reserved, Blizzard is the intellectual property owner of WoW and WoW Classic
and any data pertaining thereto

Author:
David "Gahd" Couples <gahdania@gahd.online>
"""
from battlenet_client.wow.base import *
from battlenet_client.wow.game_data import achievement, achievement_category, achievement_media

from typing import Union, Optional
from requests_oauthlib import OAuth2Session
from oic import oic


class AchievementCategory(WoWBase):
    """Defines the index achievement categories within World of Warcraft

    Args:
        region_tag (str): region abbreviation for use with the APIs

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        general (list of dict): the entire index of the categories
        root (list of dict): the index of the root categories
        guild (list of dict): the index of the guild categories

    Notes:
        This class no longer functions as it was when originally created. Blizzard changed the Index into a list of
            IDs and URLs, vice a list of the actual achievement categories (including the data)
    """

    def __init__(self, client: Union[OAuth2Session, oic.Client], category_id: Optional[int] = None, *,
                 release: Optional[str] = None, locale: Optoinal[str] = None):
        (url, params) = achievement_category(region_tag, category_id=category_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        super().__init__(region_tag, api_data['id'], release=release, locale=locale)

        self.category_id = api_data['id']
        self.name = api_data['name']
        self.achievements = api_data['achievements']
        self.alliance = api_data['aggregates_by_faction']['alliance']
        self.horde = api_data['aggregates_by_faction']['horde']
        self.is_guild_category = bool(api_data['is_guild_category'])
        self.icon = api_data['media']['key']

    def __str__(self):
        return self.__class__.__name__


class Achievement:
    """Defines the index achievements within World of Warcraft

    Args:
        region_tag (str): region abbreviation for use with the APIs

    Keyword Args:
        release (str, optional): the release to use.
        locale (str, optional): the localization to use from the API, The default of None means all localizations
        scope (list of str, optional): the scope or scopes to use during the endpoints that require the
            Web Application Flow
        redirect_uri (str, optional): the URI to return after a successful authentication between the user and Blizzard
        client_id (str): the client ID from the developer portal
        client_secret (str): the client secret from the developer portal

    Attributes:
        achievements (list of dict): the list of the achievements

    Notes:
        This class no longer functions as it was when originally created. Blizzard changed the Index into a list of
            IDs and URLs, vice a list of the achievements (including the data)
    """

    def __init__(self, client: Union[OAuth2Session, OicClient], achievement_id: Optional[str, int] = None, *,
                 release: Optional[str] = None, locale: Optional[str] = None):

        (url, params) = achievement(region_tag, achievement_id, release=release, locale=locale)
        api_data = client.get(url, params=params)

        if not achievement_id:
            self.achievements = api_data['achievements']
        else:
            self.achievement_id = api_data['id']
            self.name = api_data['name']
            self.description = api_data['description']
            self.points = int(api_data['points'])
            self.is_account_wide = bool(api_data['is_account_wide'])
            self.criteria = api_data['criteria']
            self.icon = api_data['media']['key']['href']

    def __str__(self):
        return self.__class__.__name__



"""Defines the generic Region

Classes:
    Region

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""

__version__ = '1.0.0'
__author__ = 'David \'Gahd\' Couples'


VALID_REGIONS = ('us', 'kr', 'eu', 'tw', 'cn')


class Region:

    class Tag:
        """Defines the region's abbreviation (tag)"""

        #: Region abbreviation for North America
        NORTH_AMERICA = VALID_REGIONS[0]

        #: Region abbreviation for North Europe
        EUROPE = VALID_REGIONS[2]

        #: Region abbreviation for Taiwan
        TAIWAN = VALID_REGIONS[3]

        #: Region abbreviation for Korea
        KOREA = VALID_REGIONS[1]

        #: Region abbreviation for China
        CHINA = VALID_REGIONS[4]

    class Id:
        """Defines the Region IDs for Diablo III, Starcraft 2, and Hearthstone"""

        #: Region ID for North America
        NORTH_AMERICA = 1

        #: Region ID for Europe
        EUROPE = 2

        #: Region ID for Korea and Taiwan (APAC)
        APAC = 3

        #: Region ID for China
        CHINA = 5



""" Cache interface

"""
import datetime

from typing import Optional, Union
from requests_oauthlib import OAuth2Session
from oic import oic
from requests import Response
from abc import ABC, abstractmethod

__version__ = '1.0.0'
__author__ = 'David "Gahd" Couples'


class Cache(ABC):
    """Base Cache Object"""

    def __init__(self, game: str):
        """Initializes the :obj:`Cache`"""
        self.game = game.lower()

    def __str__(self):
        return f"{__class__.__name__} {__version__}"

    def __repr__(self):
        return f"{__class__.__name__} {__version__}"

    @abstractmethod
    def select(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> bytes:
        """Gathers one row of data from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bytes): Returns the data from the cache
        """
        pass

    @abstractmethod
    def select_all(self, region: str, release: str, category: str, obj_id: Optional[int] = None):
        """Gathers all rows of data from the cache

                Args:
                    region (str): the game's region
                    release (str): the game's release name
                    category (str): the category for the data
                    obj_id (int, optional): the ID for the cached data

                Returns:
                    (bytes): Returns the data from the cache
                """
        pass

    @abstractmethod
    def delete(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> int:
        """Removed items from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (int): number of rows deleted
        """
        pass

    @abstractmethod
    def check(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> bool:
        """Returns if the data is found and if it is expired

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the data's category
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bool): True if the object is found and has not expired. False otherwise
        """
        pass

    @abstractmethod
    def submit(self, region: str, release: str, category: str,  response: Response,
               obj_id: Optional[int] = None) -> int:
        """Adds or updates data to/in the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the data's category
            response (:obj:`Response`): the response object from the API
            obj_id (int, optional): the ID for the cached data

        Returns:
            (int): number of rows inserted or updated
        """
        pass

    @abstractmethod
    def category(self, name: str):
        """Returns the category data from the cache

        Args:
            name (str): name of the category

        Returns:
            (cache dependant): the row of for the provided category
        """
        pass

    def next_update(self, last_modified: str, category_name: str):
        """Returns the next update value

        Args:
            last_modified (str): string of the last modified value from the API
            category_name (str): name of the category

        Returns:
            :obj:`datetime`: the datetime of the next update
        """
        last_modified = datetime.strptime(last_modified, '%a, %d %b %Y %H:%M:%S %Z')
        last_modified += timedelta(weeks=self.category(category_name)['duration'])
        return last_modified



from .cache import Cache
from battlenet_client.utils import localize, next_update

from pymongo import MongoClient, ReturnDocument
from datetime import datetime, timedelta
from typing import Optional, List, Dict
import json
from os import path
from requests import Response


__version__ = 'v1.0.0'


class Mongo(Cache):

    def __init__(self, game: str, db_user: str, db_pass: str,  host: Optional[str] = '127.0.0.1',
                 port: Optional[int] = 27017, *, reset: Optional[bool] = False):
        super().__init__(game)

        db_url = f"mongodb://{db_user}:{db_pass}@{host}:{port}/cache"

        self._client = MongoClient(db_url)
        self._database = self._client['cache']
        self._cache = self._database[self.game]
        self._categories = self._database['categories']

        if reset:
            self._cache.drop()
            self._categories.drop()

            categories = list()
            current_dir = path.dirname(path.realpath(__file__))
            with open(f"{current_dir}/data.txt", "r") as config:
                for line in config.readlines():
                    categories.append(json.loads(line))

            self._categories.insert_many(categories)

    def __str__(self):
        return f"{self.__class__.__name__} {__version__}"

    def __repr__(self):
        return f"{self.__class__.__name__} {__version__}"

    def select(self, region: str, release: str, category: str, obj_id: Optional[int] = None):
        """Gathers data from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bytes): Returns the data from the cache
        """
        cache_data = self._cache.find_one({"region": region, "release": release, "category": category,
                                           "object_id": obj_id}, projection={'data': True, '_id': False})

        return cache_data['data']

    def select_all(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> List[bytes]:
        cache_data = self._cache.find({"region": region, "release": release, "category": category,
                                      "object_id": obj_id}, projection={'data': true, '_id': False})

        return [x['data'] for x in cache_data]

    def delete(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> int:
        """Removed items from the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the category for the data
            obj_id (int, optional): the ID for the cached data

        Returns:
            (int): number of rows deleted
        """
        deleted_data = self._cache.find_one_and_delete({'region': region, 'release': release, 'category': category,
                                                        'object_id': obj_id})
        if deleted_data:
            return 1

        return 0

    def check(self, region: str, release: str, category: str, obj_id: Optional[int] = None) -> bool:
        """Returns if the data is found and if it is expired

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the data's category
            obj_id (int, optional): the ID for the cached data

        Returns:
            (bool): True if the object is found and has not expired. False otherwise
        """
        check_data = self._cache.find_one({"region": region, "release": release, "category": category,
                                           "object_id": obj_id,
                                           'next_update': {'$gte': datetime.datetime.now(datetime.UTC)}},
                                          projection={'next_update': True, '_id': False})

        return True if check_data else False

    def submit(self, region: str, release: str, category: str,  response: Response,
               obj_id: Optional[int] = None) -> int:
        """Adds or updates data to/in the cache

        Args:
            region (str): the game's region
            release (str): the game's release name
            category (str): the data's category
            response (:obj:`Response`): the response from the API endpoint
            obj_id (int, optional): the ID for the cached data

        Returns:
            (int): number of rows inserted or updated
        """
        submitted_document = self._cache.find_one_and_update(
            {"region": region, "release": release, "category": category, "object_id": obj_id},
            {
                "$set": {"data": data, "next_update": self.next_update(response.headers['last-modified'],
                                                                       category)},
                "$setOnInsert": {"region": region, "release": release, "category": category, "object_id": obj_id}},
            {"data": 1}, upsert=True,
            return_document=ReturnDocument.AFTER)

        return 1 if submitted_document else 0

    def category(self, name: str):
        """Returns the category data from the cache

        Args:
            name (str): name of the category

        Returns:
            (:obj:`pymongo.Row`): the row of for the provided category
        """
        return self._categories.find_one({"name": name})



"""Defines exceptions related to the Battle.net API wrappers

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""


__version__ = '1.0.0'
__author__ = 'David \'Gahd\' Couples'


class BNetError(Exception):
    """Battle.net Base Exception class
    """
    pass


class BNetRegionError(BNetError):
    """Exception raised when there is a problem with the region, when the region does exist
    """
    pass


class BNetValueError(BNetError):
    """Exception for when bad data is present"""
    pass


class BNetReleaseError(BNetError):
    """Exception raised when there is a problem with the region
    """
    pass
